@@ -109,12 +109,32 @@ class ZMQTerminalInteractiveShell(TerminalInteractiveShell):
         """
     )
 
+    use_kernel_is_complete = Bool(True, config=True,
+        help="""Whether to use the kernel's is_complete message
+        handling. If False, then the frontend will use its
+        own is_complete handler.
+        """
+    )
+    kernel_is_complete_timeout = Float(1, config=True,
+        help="""Timeout (in seconds) for giving up on a kernel's is_complete
+        response.
+
+        If the kernel does not respond at any point within this time,
+        the kernel will no longer be asked if code is complete, and the
+        console will default to the built-in is_complete test.
+        """
+    )
+
     manager = Instance('jupyter_client.KernelManager', allow_none=True)
     client = Instance('jupyter_client.KernelClient', allow_none=True)
     def _client_changed(self, name, old, new):
         self.session_id = new.session.session
     session_id = Unicode()
 
+    def __init__(self, *args, **kwargs):
+        super(ZMQTerminalInteractiveShell, self).__init__(*args, **kwargs)
+        self._source_lines_buffered = []
+
     def init_completer(self):
         """Initialize the completion machinery.
 
@@ -220,6 +240,43 @@ class ZMQTerminalInteractiveShell(TerminalInteractiveShell):
 
             self.execution_count = int(content["execution_count"] + 1)
 
+    def handle_is_complete_reply(self, msg_id, timeout=None):
+        """
+        Wait for a repsonse from the kernel, and return two values:
+            more? - (boolean) should the frontend ask for more input
+            indent - an indent string to prefix the input
+        Overloaded methods may want to examine the comeplete source. Its is
+        in the self._source_lines_buffered list.
+        """
+        ## Get the is_complete response:
+        msg = None
+        try:
+            msg = self.client.shell_channel.get_msg(block=True, timeout=timeout)
+        except Empty:
+            warn('The kernel did not respond to an is_complete_request. '
+                 'Setting `use_kernel_is_complete` to False.')
+            self.use_kernel_is_complete = False
+            return False, ""
+        ## Handle response:
+        if msg["parent_header"].get("msg_id", None) != msg_id:
+            warn('The kernel did not respond properly to an is_complete_request: %s.' % str(msg))
+            return False, ""
+        else:
+            status = msg["content"].get("status", None)
+            indent = msg["content"].get("indent", "")
+        ## Return more? and indent string
+        if status == "complete":
+            return False, indent
+        elif status == "incomplete":
+            return True, indent
+        elif status == "invalid":
+            raise SyntaxError()
+        elif status == "unknown":
+            return False, indent
+        else:
+            warn('The kernel sent an invalid is_complete_reply status: "%s".' % status)
+            return False, indent
+
     include_other_output = Bool(False, config=True,
         help="""Whether to include output from clients
         other than this one sharing the same kernel.
@@ -396,6 +453,9 @@ class ZMQTerminalInteractiveShell(TerminalInteractiveShell):
             signal.signal(signal.SIGINT, double_int)
             content = req['content']
             read = getpass if content.get('password', False) else input
+            if self.has_readline:
+                # Disable tab completion while we prompt for arbitrary input
+                self.readline.set_completer(None)
             try:
                 raw_data = read(content["prompt"])
             except EOFError:
@@ -407,6 +467,9 @@ class ZMQTerminalInteractiveShell(TerminalInteractiveShell):
             finally:
                 # restore SIGINT handler
                 signal.signal(signal.SIGINT, real_handler)
+                # Restore tab completion ready for our next input prompt
+                if self.has_readline:
+                    self.set_readline_completer()
 
             # only send stdin reply if there *was not* another request
             # or execution finished while we were reading.
@@ -476,6 +539,11 @@ class ZMQTerminalInteractiveShell(TerminalInteractiveShell):
                 return False
         return True
 
+    _next_line_indent = ''
+
+    def _indent_current_str(self):
+        return self._next_line_indent
+
     def interact(self, display_banner=None):
         """Closely emulate the interactive Python console."""
 
@@ -502,6 +570,7 @@ class ZMQTerminalInteractiveShell(TerminalInteractiveShell):
 
         if self.has_readline:
             self.readline_startup_hook(self.pre_readline)
+            self.refill_readline_hist()
             hlen_b4_cell = self.readline.get_current_history_length()
         else:
             hlen_b4_cell = 0
@@ -534,6 +603,7 @@ class ZMQTerminalInteractiveShell(TerminalInteractiveShell):
                         self.rl_do_indent = True
 
                 else:
+                    self._next_line_indent = ''
                     try:
                         prompt = self.separate_in + self.prompt_manager.render('in')
                     except Exception:
@@ -550,7 +620,7 @@ class ZMQTerminalInteractiveShell(TerminalInteractiveShell):
                 #double-guard against keyboardinterrupts during kbdint handling
                 try:
                     self.write('\n' + self.get_exception_only())
-                    source_raw = self.input_splitter.raw_reset()
+                    source_raw = self.get_source_and_reset()
                     hlen_b4_cell = self._replace_rlhist_multiline(source_raw, hlen_b4_cell)
                     more = False
                 except KeyboardInterrupt:
@@ -573,8 +643,7 @@ class ZMQTerminalInteractiveShell(TerminalInteractiveShell):
                 self.showtraceback()
             else:
                 try:
-                    self.input_splitter.push(line)
-                    more = self.input_splitter.push_accepts_more()
+                    more, indent = self.handle_source(line)
                 except SyntaxError:
                     # Run the code directly - run_cell takes care of displaying
                     # the exception.
@@ -582,8 +651,10 @@ class ZMQTerminalInteractiveShell(TerminalInteractiveShell):
                 if (self.SyntaxTB.last_syntax_error and
                     self.autoedit_syntax):
                     self.edit_syntax_error()
+                if more:
+                    self._next_line_indent = indent
                 if not more:
-                    source_raw = self.input_splitter.raw_reset()
+                    source_raw = self.get_source_and_reset()
                     hlen_b4_cell = self._replace_rlhist_multiline(source_raw, hlen_b4_cell)
                     self.run_cell(source_raw)
 
@@ -596,3 +667,27 @@ class ZMQTerminalInteractiveShell(TerminalInteractiveShell):
         self.history_manager = ZMQHistoryManager(client=self.client)
         self.configurables.append(self.history_manager)
 
+    def get_source_and_reset(self):
+        """
+        Get the source code entered so far, and reset source input line(s).
+        """
+        source, self._source_lines_buffered = self._source_lines_buffered, []
+        return "\n".join(source)
+
+    def handle_source(self, line):
+        """
+        Save the line with previous source lines, and see if more
+        input is needed. Returns more? and indent_string.
+        """
+        self._source_lines_buffered.append(line)
+        ## Ask client if line is complete; get indent for next line:
+        if self.use_kernel_is_complete:
+            msg_id = self.client.is_complete("\n".join(self._source_lines_buffered))
+            return self.handle_is_complete_reply(msg_id, timeout=self.kernel_is_complete_timeout)
+        else:
+            more = (line != "")
+            return more, ""
+
+    def init_virtualenv(self):
+        # No need to do this in the frontend, and the warning is confusing
+        pass
